unit BinaryIon;

{$IFDEF FPC}
  {$MODE Delphi}
{$ENDIF}

interface

uses Classes, SysUtils, IonSymbols, Variants, Math, StrUtils;

const
  TID_NULL = 0;
  TID_BOOLEAN = 1;
  TID_POSINT = 2;
  TID_NEGINT = 3;
  TID_FLOAT = 4;
  TID_DECIMAL = 5;
  TID_TIMESTAMP = 6;
  TID_SYMBOL = 7;
  TID_STRING = 8;
  TID_CLOB = 9;
  TID_BLOB = $A;
  TID_LIST = $B;
  TID_SEXP = $C;
  TID_STRUCT = $D;
  TID_TYPEDECL = $E;
  TID_UNUSED = $F;

type
  EIonException = class(Exception);
  EIonUnexpectedEOF = class(EIonException)
  public
    constructor Create;
  end;

  TBinaryIonParser = class
  private
    type TParserState = (sInvalid, sBeforeField, sBeforeTID, sBeforeValue, sAfterValue, sEOF);
    type TContainerRec = record
      NextPos: Int64;
      TID, Remaining: Integer;
    end;
  private
    FStream: TStream;
    FInitPos: Int64;
    FLocalRemaining: Integer;
    FEOF: Boolean;
    FState: TParserState;
    FNeedHasNext: Boolean;

    FIsInStruct: Boolean;
    FContainerStack: array of TContainerRec;

    FValueTID, FValueFieldID, FParentTID: Integer;
    FValueLen: Cardinal;
    FValueIsNull, FValueIsTrue: Boolean;
    FValue: Variant;

    FAnnotations: array of Integer;
    FSymbols: TSymbolTable;
    FCatalog: array of TIonCatalogItem;
    FDidImports: Boolean;

    procedure HasNextRaw;

    function Read: Byte; overload;
    procedure Read(var Buf; Count: Integer); overload;
    function ReadVarInt: Integer;
    function ReadVarUInt: Cardinal;
    function ReadDecimal: Extended;
    function ReadTypeID: Integer;
    function ReadFieldID: Integer;
    procedure Skip(const Count: Int64);

    procedure Push(TypeID: Integer; const NextPosition: Int64; NextRemaining: Integer);
    procedure Pop;

    procedure ParseSymbolTable;
    procedure GatherImports;
    procedure ReadImport;
    function FindCatalogItem(const Name: string): TIonCatalogItem;

    procedure ClearValue;
    procedure PrepareValue;
    procedure LoadScalarValue;

    procedure LoadAnnotations;
    procedure CheckVersionMarker;
    procedure Fault(const Msg: string);
  public
    constructor Create(AStream: TStream);
    destructor Destroy; override;

    procedure AddToCatalog(const AName: string; AVersion: Integer; ASymbols: TStringList);
    procedure ForceImport(ASymbols: TStringList);

    function HasNext: Boolean;
    function Next: Integer;
    procedure Reset;

    function GetFieldName: string;
    function GetFieldNameSymbol: TSymbolToken;
    function GetTypeName: string;

    function IntValue: Integer;
    function StringValue: string;
    function SymbolValue: string;
    function LOBValue: TBytes;
    function DecimalValue: string;

    procedure StepIn;
    procedure StepOut;

    procedure Print(List: TStrings);

    property IsInStruct: Boolean read FIsInStruct;
  end;

implementation

{ TIonParser }

const
  LEN_IS_VAR_LEN = $E;
  LEN_IS_NULL = $F;

  VERSION_MARKER: array[0..2] of Byte = ($01, $00, $EA);

constructor TBinaryIonParser.Create(AStream: TStream);
begin
  FStream := AStream;
  FInitPos := AStream.Position;
  Reset;
  FSymbols := TSymbolTable.Create;
end;

destructor TBinaryIonParser.Destroy;
begin
  FSymbols.Free;

  inherited;
end;

procedure TBinaryIonParser.Reset;
begin
  FState := sBeforeTID;
  FNeedHasNext := True;
  FLocalRemaining := -1;
  FEOF := False;
  FIsInStruct := False;
  FContainerStack := nil;
  FStream.Position := FInitPos;
end;

procedure TBinaryIonParser.AddToCatalog(const AName: string; AVersion: Integer; ASymbols: TStringList);
begin
  SetLength(FCatalog, Length(FCatalog) + 1);
  FCatalog[High(FCatalog)] := TIonCatalogItem.Create(AName, AVersion, ASymbols);
end;

function TBinaryIonParser.HasNext: Boolean;
var
  i: Integer;
begin
  while FNeedHasNext and not FEOF do
  begin
    HasNextRaw;
    // Check if we need to skip system data in the root (version marker and symbol table)
    if (Length(FContainerStack) = 0) and not FValueIsNull then
    begin
      if FValueTID = TID_SYMBOL then
      begin
        if FValue = SID_ION_1_0 then
          FNeedHasNext := True;
      end
      else if FValueTID = TID_STRUCT then
      begin
        for i := 0 to High(FAnnotations) do
          if FAnnotations[i] = SID_ION_SYMBOL_TABLE then
          begin
            ParseSymbolTable;
            FNeedHasNext := True;
            Break;
          end;
      end;
    end;
  end;

  Result := not FEOF;
end;

procedure TBinaryIonParser.HasNextRaw;
begin
  ClearValue;
  while (FValueTID = -1) and not FEOF do
  begin
    FNeedHasNext := False;
    case FState of
      sBeforeField:
      begin
        if FValueFieldID <> UNKNOWN_SYMBOL_ID then
          Fault('FieldID already set');

        FValueFieldID := ReadFieldID;
        if FValueFieldID <> UNKNOWN_SYMBOL_ID then
          FState := sBeforeTID
        else
          FEOF := True;
      end;

      sBeforeTID:
      begin
        FState := sBeforeValue; // Changed by ReadTypeID in certain cases
        FValueTID := ReadTypeID;
        if FValueTID = -1 then
        begin
          FState := sEOF;
          FEOF := True;
          Break;
        end;

        if FValueTID = TID_TYPEDECL then
        begin
          if FValueLen = 0 then
            CheckVersionMarker
          else
            LoadAnnotations;
        end;
      end;

      sBeforeValue:
      begin
        Skip(FValueLen);
        FState := sAfterValue;
      end;

      sAfterValue:
      begin
        if FIsInStruct then
          FState := sBeforeField
        else
          FState := sBeforeTID;
      end;

      sEOF: ;
      else Fault('Internal error');
    end;
  end;
end;

function TBinaryIonParser.Next: Integer;
begin
  if HasNext then
  begin
    FNeedHasNext := True;
    Result := FValueTID;
  end
  else
    Result := -1;
end;

procedure TBinaryIonParser.Push(TypeID: Integer; const NextPosition: Int64; NextRemaining: Integer);
begin
  SetLength(FContainerStack, Length(FContainerStack) + 1);
  with FContainerStack[High(FContainerStack)] do
  begin
    NextPos := NextPosition;
    TID := TypeID;
    Remaining := NextRemaining;
  end;
end;

procedure TBinaryIonParser.Pop;
begin
  if Length(FContainerStack) = 0 then
    raise EIonException.Create('Cannot pop');

  SetLength(FContainerStack, Length(FContainerStack) - 1);
end;

procedure TBinaryIonParser.StepIn;
var
  NextRem: Integer;
begin
  if not (FValueTID in [TID_STRUCT, TID_LIST, TID_SEXP]) or (FEOF) then
    Fault('Cannot step in');

  if (FValueIsNull and (FState <> sAfterValue)) or ((not FValueIsNull) and (FState <> sBeforeValue))  then
    Fault('Invalid state in StepIn');

  NextRem := FLocalRemaining;
  if NextRem <> -1 then
  begin
    Dec(NextRem, FValueLen);
    if NextRem < 0 then
      NextRem := 0;
  end;
  Push(FParentTID, FStream.Position + FValueLen, NextRem);

  FIsInStruct := FValueTID = TID_STRUCT;
  if FIsInStruct then
    FState := sBeforeField
  else
    FState := sBeforeTID;

  FLocalRemaining := FValueLen;
  FParentTID := FValueTID;
  ClearValue;
  FNeedHasNext := True;
end;

procedure TBinaryIonParser.StepOut;
var
  Rec: TContainerRec;
  CurPos: Int64;
begin
  if Length(FContainerStack) = 0 then
    Fault('Cannot step out');

  Rec := FContainerStack[High(FContainerStack)];
  Pop;

  FEOF := False;
  FParentTID := Rec.TID;
  if FParentTID = TID_STRUCT then
  begin
    FIsInStruct := True;
    FState := sBeforeField;
  end
  else
  begin
    FIsInStruct := False;
    FState := sBeforeTID;
  end;
  FNeedHasNext := True;

  ClearValue;
  CurPos := FStream.Position;
  if Rec.NextPos > CurPos then
  begin
    Skip(Rec.NextPos - CurPos);
  end
  else if Rec.NextPos < CurPos then
    raise EIonException.CreateFmt('Invalid position in StepOut (CurPos: %d; NextPos: %d)', [CurPos, Rec.NextPos]);

  FLocalRemaining := Rec.Remaining;
end;

function TBinaryIonParser.Read: Byte;
begin
  if FLocalRemaining <> -1 then
  begin
    if FLocalRemaining < 1 then
      raise EIonUnexpectedEOF.Create;
    Dec(FLocalRemaining);
  end;

  if FStream.Read(Result, 1) <> 1 then
    raise EIonUnexpectedEOF.Create;
end;

procedure TBinaryIonParser.Read(var Buf; Count: Integer);
begin
  if FLocalRemaining <> -1 then
  begin
    Dec(FLocalRemaining, Count);
    if FLocalRemaining < 0 then
      raise EIonUnexpectedEOF.Create;
  end;
  if FStream.Read(Buf, Count) <> Count then
    raise EIonUnexpectedEOF.Create;
end;

function TBinaryIonParser.ReadFieldID: Integer;
begin
  if (FLocalRemaining <> -1) and (FLocalRemaining < 1) then
    Exit(-1);

  try
    Result := ReadVarUInt;
  except
    on E: EIonUnexpectedEOF do
      Result := -1
  end;
end;

function TBinaryIonParser.ReadTypeID: Integer;
var
  B: Byte;
  Len: Cardinal;
begin
  if FLocalRemaining <> -1 then
  begin
    if FLocalRemaining < 1 then
      Exit(-1);
    Dec(FLocalRemaining);
  end;

  if FStream.Read(B, 1) <> 1 then
    Exit(-1);

  Result := B shr 4;
  Len := B and $F;

  if Len = LEN_IS_VAR_LEN then
  begin
    Len := ReadVarUInt;
  end
  else if Len = LEN_IS_NULL then
  begin
    Len := 0;
    FState := sAfterValue;
  end
  else if Result = TID_NULL then
  begin
    // Must have LEN_IS_NULL
    Fault('Invalid null type descriptor');
  end
  else if Result = TID_BOOLEAN then
  begin
    if Len > 1 then
      Fault('Invalid boolean value');
    FValueIsTrue := Len = 1;
    Len := 0;
    FState := sAfterValue;
  end
  else if Result = TID_STRUCT then
  begin
    if Len = 1 then
      Len := ReadVarUInt;
  end;

  FValueLen := Len;
end;

{$HINTS OFF} // it doesn't like my finally construct...
function TBinaryIonParser.ReadVarInt: Integer;
var
  B: Byte;
  Negative: Boolean;
begin
  B := Read();
  Negative := (B and $40) <> 0;
  Result := (B and $3F);
  try
    if B and $80 <> 0 then
      Exit;

    B := Read();
    Result := (Result shl 7) or (B and $7F);
    if B and $80 <> 0 then
      Exit;
    B := Read();
    Result := (Result shl 7) or (B and $7F);
    if B and $80 <> 0 then
      Exit;
    B := Read();
    Result := (Result shl 7) or (B and $7F);
    if B and $80 <> 0 then
      Exit;
    B := Read();
    Result := (Result shl 7) or (B and $7F);
    if B and $80 <> 0 then
      Exit;
  finally
    if Negative then
      Result := -Result;
  end;

  raise EIonException.Create('Int overflow');
end;
{$HINTS ON}

function TBinaryIonParser.ReadVarUInt: Cardinal;
var
  B: Byte;
begin
  B := Read();
  Result := (B and $7F);
  if B and $80 <> 0 then
    Exit;
  B := Read();
  Result := (Result shl 7) or (B and $7F);
  if B and $80 <> 0 then
    Exit;
  B := Read();
  Result := (Result shl 7) or (B and $7F);
  if B and $80 <> 0 then
    Exit;
  B := Read();
  Result := (Result shl 7) or (B and $7F);
  if B and $80 <> 0 then
    Exit;
  B := Read();
  Result := (Result shl 7) or (B and $7F);
  if B and $80 <> 0 then
    Exit;

  raise EIonException.Create('Int overflow');
end;

function TBinaryIonParser.ReadDecimal: Extended;
var
  Exponent, Rem, i, j: Integer;
  B: TBytes;
  V: UInt64;
  VB: array[0..7] of Byte absolute V;
  Signed: Boolean;
begin
  if FValueLen = 0 then
  begin
    Result := 0;
  end
  else
  begin
    Rem := FLocalRemaining - Integer(FValueLen);
    FLocalRemaining := FValueLen;
    Exponent := ReadVarInt;

    if FLocalRemaining > 0 then
    begin
      if FLocalRemaining > 8 then
        raise Exception.Create('Decimal overflow');

      Signed := False;
      SetLength(B, FLocalRemaining);
      Read(B[0], FLocalRemaining);
      if (B[0] and $80) <> 0 then
      begin
        B[0] := B[0] and $7F;
        Signed := True;
      end;
      j := 0;
      V := 0;
      for i := High(B) downto 0 do
      begin
        VB[i] := B[j];
        Inc(j);
      end;

      Result := V;
      Result := Result * Power(10, Exponent);
      if Signed then
        Result := -Result;
    end
    else
      raise ENotSupportedException.Create('Only exponent in ReadDecimal');

    FLocalRemaining := Rem;
  end;
end;

procedure TBinaryIonParser.Skip(const Count: Int64);
begin
  if FLocalRemaining <> -1 then
  begin
    Dec(FLocalRemaining, Count);
    if FLocalRemaining < 0 then
      raise EIonUnexpectedEOF.Create;
  end;

  FStream.Seek(Count, soCurrent);
end;

procedure TBinaryIonParser.ParseSymbolTable;
var
  FieldType: Integer;
begin
  Next; // shouldn't do anything

  if FValueTID <> TID_STRUCT then
    Fault('Struct expected in ParseSymbolTable');

  if FDidImports then
    Exit;

  StepIn;

  FieldType := Next;
  while FieldType <> -1 do
  begin
    if not FValueIsNull then
    begin
      case FValueFieldID of
        SID_IMPORTS:
        begin
          if FieldType = TID_LIST then
          begin
            GatherImports;
          end;
        end;

        else Fault('Unsupported symbol table field id');
      end;
    end;

    FieldType := Next;
  end;

  StepOut;
  FDidImports := True;
end;

procedure TBinaryIonParser.GatherImports;
var
  T: Integer;
begin
  StepIn;

  T := Next;
  while T <> -1 do
  begin
    if (not FValueIsNull) and (T = TID_STRUCT) then
    begin
      ReadImport;
    end;

    T := Next;
  end;

  StepOut;
end;

procedure TBinaryIonParser.ReadImport;
var
  T, Version, MaxID: Integer;
  Name: string;
  Table: TIonCatalogItem;
begin
  Version := -1;
  MaxID := -1;
  Name := '';

  StepIn;

  T := Next;
  while T <> -1 do
  begin
    if (not FValueIsNull) and (FValueFieldID <> UNKNOWN_SYMBOL_ID) then
    begin
      case FValueFieldID of
        SID_NAME: Name := StringValue;
        SID_VERSION: Version := IntValue;
        SID_MAX_ID: MaxID := IntValue;
      end;
    end;

    T := Next;
  end;

  StepOut;

  if (Name = '') or (Name = TSystemSymbols.ION) then
    Exit;

  if Version < 1 then
    Version := 1;

  Table := FindCatalogItem(Name);
  if MaxID < 0 then
  begin
    if (Table = nil) or (Version <> Table.Version) then
      raise EIonException.Create('Import ' + Name + ' lacks maxid');

    MaxID := Table.SymNames.Count;
  end;

  if Table <> nil then
    FSymbols.Import(Table, Min(MaxID, Table.SymNames.Count))
  else
    FSymbols.ImportUnknown(Name, MaxID);
end;

function TBinaryIonParser.IntValue: Integer;
var
  V: Int64;
begin
  if not (FValueTID in [TID_POSINT, TID_NEGINT]) then
    Fault('Not an int');

  PrepareValue;
  V := FValue;
  Result := V;
end;

function TBinaryIonParser.StringValue: string;
begin
  if FValueTID <> TID_STRING then
    Fault('Not a string');

  if FValueIsNull then
    Exit('');

  PrepareValue;
  Result := FValue;
end;

function TBinaryIonParser.SymbolValue: string;
begin
  if FValueTID <> TID_SYMBOL then
    Fault('Not a symbol');

  PrepareValue;
  Result := FSymbols.FindByID(FValue);
  if Result = '' then
    Result := 'SYMBOL#' + IntToStr(FValue);
end;

function TBinaryIonParser.LOBValue: TBytes;
begin
  if not (FValueTID in [TID_CLOB, TID_BLOB]) then
    Fault('Not a LOB type: ' + GetFieldName);

  if FValueIsNull then
    Exit(nil);

  SetLength(Result, FValueLen);
  Read(Result[0], FValueLen);
  FState := sAfterValue;
end;

function TBinaryIonParser.DecimalValue: string;
begin
  if FValueTID <> TID_DECIMAL then
    Fault('Not a decimal');

  PrepareValue;
  Result := FValue;
end;

procedure TBinaryIonParser.PrepareValue;
begin
  if VarIsEmpty(FValue) then
  begin
    LoadScalarValue;
  end;
end;

procedure TBinaryIonParser.LoadScalarValue;
var
  B: TBytes;
  v, i: Cardinal;
begin
  if not (FValueTID in [TID_NULL, TID_BOOLEAN, TID_POSINT, TID_NEGINT, TID_FLOAT, TID_DECIMAL, TID_TIMESTAMP, TID_SYMBOL, TID_STRING]) then
    Exit;

  if FValueIsNull then
  begin
    FValue := Variants.Null;
    Exit;
  end;

  case FValueTID of
    TID_STRING:
    begin
      SetLength(B, FValueLen);
      Read(B[0], FValueLen);
      FValue := TEncoding.UTF8.GetString(B);
    end;

    TID_POSINT, TID_NEGINT, TID_SYMBOL:
    begin
      if FValueLen = 0 then
      begin
        FValue := 0;
      end
      else if FValueLen <= 4 then
      begin
        v := 0;
        for i := FValueLen - 1 downto 0 do
          v := v or (Read shl (i*8));

        if FValueTID = TID_NEGINT then
          FValue := -Integer(v)
        else
          FValue := v;
      end
      else
        raise ENotSupportedException.CreateFmt('Int too long: %d', [FValueLen]);
    end;

    TID_DECIMAL:
    begin
      FValue := ReadDecimal;
    end;
  end;

  FState := sAfterValue;
end;

procedure TBinaryIonParser.ClearValue;
begin
  FValueTID := -1;
  VarClear(FValue);
  FValueIsNull := False;
  FValueFieldID := UNKNOWN_SYMBOL_ID;
  FAnnotations := nil;
end;

procedure TBinaryIonParser.LoadAnnotations;
var
  Len: Integer;
  MaxPos: Int64;
begin
  Len := ReadVarUInt;
  MaxPos := FStream.Position + Len;
  while FStream.Position < MaxPos do
  begin
    SetLength(FAnnotations, Length(FAnnotations) + 1);
    FAnnotations[High(FAnnotations)] := ReadVarUInt;
  end;
  FValueTID := ReadTypeID;
end;

procedure TBinaryIonParser.CheckVersionMarker;
var
  i: Integer;
begin
  for i := 0 to High(VERSION_MARKER) do
    if Read() <> VERSION_MARKER[i] then
      Fault('Unknown version marker');

  FValueLen := 0;
  FValueTID := TID_SYMBOL;
  FValue := SID_ION_1_0;
  FValueIsNull := False;
  FValueFieldID := UNKNOWN_SYMBOL_ID;
  FState := sAfterValue;
end;

procedure TBinaryIonParser.Fault(const Msg: string);
begin
  raise EIonException.CreateFmt('%s at %d', [Msg, FStream.Position - 1]);
end;

function TBinaryIonParser.FindCatalogItem(const Name: string): TIonCatalogItem;
begin
  for Result in FCatalog do
    if Result.Name = Name then
      Exit;
  Result := nil;
end;

procedure TBinaryIonParser.ForceImport(ASymbols: TStringList);
var
  Item: TIonCatalogItem;
begin
  Item := TIonCatalogItem.Create('Forced', 1, ASymbols);
  try
    FSymbols.Import(Item, ASymbols.Count);
  finally
    Item.Free;
  end;
end;

function TBinaryIonParser.GetFieldName: string;
begin
  if FValueFieldID = UNKNOWN_SYMBOL_ID then
    Result := ''
  else
    Result := FSymbols.FindByID(FValueFieldID);
end;

function TBinaryIonParser.GetFieldNameSymbol: TSymbolToken;
begin
  if FValueFieldID = UNKNOWN_SYMBOL_ID then
    Exit(TSymbolToken.Create('', UNKNOWN_SYMBOL_ID));

  Result := TSymbolToken.Create(FSymbols.FindByID(FValueFieldID), FValueFieldID);
end;

function TBinaryIonParser.GetTypeName: string;
begin
  if Length(FAnnotations) = 0 then
    Exit('');

  Result := FSymbols.FindByID(FAnnotations[0]);
end;

procedure TBinaryIonParser.Print(List: TStrings);
var
  Indent: string;

  function PrintLOB(const B: TBytes): string;
  var
    i: Integer;
  begin
    if B = nil then
      Exit('null');

    Result := '';
    for i := 0 to High(B) do
      Result := Result + IntToHex(B[i], 2) + ' ';
    Delete(Result, Length(Result), 1);
  end;

  procedure IonWalk(SuperT: Integer);
  var
    L, TN: string;
    T: Integer;
  begin
    while HasNext do
    begin
      if SuperT = TID_STRUCT then
        L := GetFieldName + ':'
      else
        L := '';

      T := Next;
      if T in [TID_STRUCT, TID_LIST] then
      begin
        if L <> '' then
          List.Add(Indent + L);
        L := GetTypeName;
        if L <> '' then
          List.Add(Indent + L + '::');
        List.Add(Indent + StrUtils.IfThen(T = TID_STRUCT, '{', '['));
        Indent := Indent + '  ';
        StepIn;
        IonWalk(T);
        StepOut;
        Delete(Indent, Length(Indent) - 1, 2);
        List.Add(Indent + StrUtils.IfThen(T = TID_STRUCT, '}', ']'));
      end
      else
      begin
        if T = TID_STRING then
          L := L + '"' + StringValue + '"'
        else if T in [TID_CLOB, TID_BLOB] then
          L := L + '{' + PrintLOB(LOBValue) + '}'
        else if T = TID_POSINT then
          L := L + IntToStr(IntValue)
        else if T = TID_SYMBOL then
        begin
          TN := GetTypeName;
          if TN <> '' then
            TN := TN + '::';
          L := L + TN + SymbolValue;
        end
        else if T = TID_DECIMAL then
          L := L + DecimalValue
        else
          L := L + 'TID ' + IntToStr(T);
        List.Add(Indent + L);
      end;
    end;
  end;

begin
  Reset;
  Indent := '';
  IonWalk(-1);
end;

{ EIonUnexpectedEOF }

constructor EIonUnexpectedEOF.Create;
begin
  inherited Create('Unexpected EOF');
end;

end.
