unit DrmIon;

{$IFDEF FPC}
  {$MODE Delphi}
{$ENDIF}

interface

uses SysUtils, Classes, BinaryIon, DECCipher, DECHash, Crypto,
     ULZMADecoder, ULZMACommon;

type
  TDrmIonVoucher = class
  private
    FEnvelope, FVoucher, FDRMKey: TBinaryIonParser;
    FEncAlgorithm, FEncTransformation, FHashAlgorithm: string;
    FLockParams: TStringList;
    FCipherText, FCipherIV, FSecretKey: TBytes;

    procedure ParseVoucher;
  public
    constructor Create(AVoucherEnv: TStream);

    procedure DecryptVoucher;

    procedure Parse;
    procedure PrintEnvelope(List: TStrings);
    procedure PrintVoucher(List: TStrings);
    procedure PrintKey(List: TStrings);

    property SecretKey: TBytes read FSecretKey;
  end;

  TOnVoucherRequired = function(const Name: string): TDrmIonVoucher of object;

  TDrmIon = class
  private
    FIon: TBinaryIonParser;
    FVoucher: TDrmIonVoucher;
    FVoucherName: string;
    FKey: TBytes;
    FOnVoucherRequired: TOnVoucherRequired;

    procedure ProcessPage(var CT, CIV: TBytes; OutPages: TStream);
  public
    constructor Create(AIonStream: TStream; AOnVoucherRequired: TOnVoucherRequired);
    destructor Destroy; override;

    procedure Parse(const OutFN: string);
    procedure Print(List: TStrings);
  end;

implementation

uses Kinf;

const
  SYM_NAMES: array[0..64] of string = (
    'com.amazon.drm.Envelope@1.0', 'com.amazon.drm.EnvelopeMetadata@1.0',
    'size', 'page_size', 'encryption_key', 'encryption_transformation',
    'encryption_voucher', 'signing_key', 'signing_algorithm', 'signing_voucher',
    'com.amazon.drm.EncryptedPage@1.0', 'cipher_text', 'cipher_iv',
    'com.amazon.drm.Signature@1.0', 'data', 'com.amazon.drm.EnvelopeIndexTable@1.0',
    'length', 'offset', 'algorithm', 'encoded', 'encryption_algorithm',
    'hashing_algorithm', 'expires', 'format', 'id', 'lock_parameters',
    'strategy', 'com.amazon.drm.Key@1.0', 'com.amazon.drm.KeySet@1.0',
    'com.amazon.drm.PIDv3@1.0', 'com.amazon.drm.PlainTextPage@1.0',
    'com.amazon.drm.PlainText@1.0', 'com.amazon.drm.PrivateKey@1.0',
    'com.amazon.drm.PublicKey@1.0', 'com.amazon.drm.SecretKey@1.0',
    'com.amazon.drm.Voucher@1.0', 'public_key', 'private_key',
    'com.amazon.drm.KeyPair@1.0', 'com.amazon.drm.ProtectedData@1.0',
    'doctype', 'com.amazon.drm.EnvelopeIndexTableOffset@1.0', 'enddoc',
    'license_type', 'license', 'watermark', 'key', 'value',
    'com.amazon.drm.License@1.0', 'category', 'metadata', 'categorized_metadata',
    'com.amazon.drm.CategorizedMetadata@1.0', 'com.amazon.drm.VoucherEnvelope@1.0',
    'mac', 'voucher', 'com.amazon.drm.ProtectedData@2.0', 'com.amazon.drm.Envelope@2.0',
    'com.amazon.drm.EnvelopeMetadata@2.0', 'com.amazon.drm.EncryptedPage@2.0',
    'com.amazon.drm.PlainText@2.0', 'compression_algorithm',
    'com.amazon.drm.Compressed@1.0', 'priority', 'refines'
  );

procedure AddProtTable(Ion: TBinaryIonParser);
var
  Syms: TStringList;
  i: Integer;
begin
  Syms := TStringList.Create;
  for i := 0 to High(SYM_NAMES) do
    Syms.Add(SYM_NAMES[i]);

  Ion.AddToCatalog('ProtectedData', 1, Syms);
end;

{ TDrmIon }

constructor TDrmIonVoucher.Create(AVoucherEnv: TStream);
begin
  FEnvelope := TBinaryIonParser.Create(AVoucherEnv);
  AddProtTable(FEnvelope);
end;

procedure TDrmIonVoucher.DecryptVoucher;
var
  Path, Shared, Param: string;
  Kinf: TKinfStorage;
  Key: RawByteString;
  B, SharedSecret: TBytes;
  i: Integer;
begin
  Shared := 'PIDv3' + FEncAlgorithm + FEncTransformation + FHashAlgorithm;

  Path := GetKinfPath;
  if (Path = '') or not FileExists(Path) then
    raise Exception.Create('Unable to locate secure storage (tried looking at ' + Path + ')');

  Kinf := ReadKinf(Path);
  FLockParams.Sort;
  for i := 0 to FLockParams.Count - 1 do
  begin
    Param := FLockParams[i];
    if Param = 'ACCOUNT_SECRET' then
      Shared := Shared + Param + Kinf['kindle.account.tokens']
    else if Param = 'CLIENT_ID' then
      Shared := Shared + Param + Kinf['DSN']
    else
      raise Exception.Create('Unknown lock parameter: ' + Param);
  end;

  SharedSecret := TEncoding.ASCII.GetBytes(Shared);

  Hmac(THash_SHA256, SharedSecret[0], 5, @SharedSecret[0], Length(SharedSecret), Key);
  with TCipher_Rijndael.Create do
  begin
    Mode := cmCBCx;
    Init(Key[1], 32, FCipherIV[0], 16);
    SetLength(B, Length(FCipherText));
    Decode(FCipherText[0], B[0], Length(FCipherText));
    Free;
  end;
  RemovePKCS5Padding(B);

  FDRMKey := TBinaryIonParser.Create(TBytesStream.Create(B));
  AddProtTable(FDRMKey);

  if (not FDRMKey.HasNext) or (FDRMKey.Next <> TID_LIST) or (FDRMKey.GetTypeName <> 'com.amazon.drm.KeySet@1.0') then
    raise Exception.Create('Expected KeySet, got ' + FDRMKey.GetTypeName);

  FDRMKey.StepIn;
  while FDRMKey.HasNext do
  begin
    FDRMKey.Next;
    if FDRMKey.GetTypeName = 'com.amazon.drm.SecretKey@1.0' then
    begin
      FDRMKey.StepIn;
      while FDRMKey.HasNext do
      begin
        FDRMKey.Next;
        if FDRMKey.GetFieldName = 'algorithm' then
        begin
          if FDRMKey.StringValue <> 'AES' then
            raise Exception.Create('Unknown cipher algorithm: ' + FDRMKey.StringValue);
        end
        else if FDRMKey.GetFieldName = 'format' then
        begin
          if FDRMKey.StringValue <> 'RAW' then
            raise Exception.Create('Unknown key format: ' + FDRMKey.StringValue);
        end
        else if FDRMKey.GetFieldName = 'encoded' then
          FSecretKey := FDRMKey.LOBValue;
      end;
      FDRMKey.StepOut;
      Break;
    end;
  end;
  FDRMKey.StepOut;
end;

procedure TDrmIonVoucher.Parse;
var
  Field: string;
begin
  FEnvelope.Reset;
  if not FEnvelope.HasNext then
    raise Exception.Create('Envelope is empty');

  if (FEnvelope.Next <> TID_STRUCT) or (FEnvelope.GetTypeName <> 'com.amazon.drm.VoucherEnvelope@1.0') then
    raise Exception.Create('Unknown type encountered in envelope, expected VoucherEnvelope');

  FEnvelope.StepIn;
  while FEnvelope.HasNext do
  begin
    FEnvelope.Next;
    Field := FEnvelope.GetFieldName;
    if Field = 'voucher' then
    begin
      FVoucher := TBinaryIonParser.Create(TBytesStream.Create(FEnvelope.LOBValue));
      AddProtTable(FVoucher);
    end
    else if Field = 'strategy' then
    begin
      if FEnvelope.GetTypeName <> 'com.amazon.drm.PIDv3@1.0' then
        raise Exception.Create('Unknown strategy: ' + FEnvelope.GetTypeName);

      FEnvelope.StepIn;
      while FEnvelope.HasNext do
      begin
        FEnvelope.Next;
        Field := FEnvelope.GetFieldName;
        if Field = 'encryption_algorithm' then
          FEncAlgorithm := FEnvelope.StringValue
        else if Field = 'encryption_transformation' then
          FEncTransformation := FEnvelope.StringValue
        else if Field = 'hashing_algorithm' then
          FHashAlgorithm := FEnvelope.StringValue
        else if Field = 'lock_parameters' then
        begin
          FEnvelope.StepIn;
          FLockParams := TStringList.Create;
          while FEnvelope.HasNext do
          begin
            if FEnvelope.Next <> TID_STRING then
              raise Exception.Create('Expected string list for lock_parameters');
            FLockParams.Add(FEnvelope.StringValue);
          end;
          FEnvelope.StepOut;
        end;
      end;
      FEnvelope.StepOut;
    end;
  end;
  ParseVoucher;
end;

procedure TDrmIonVoucher.ParseVoucher;
begin
  if not FVoucher.HasNext then
    raise Exception.Create('Voucher is empty');

  if (FVoucher.Next <> TID_STRUCT) or (FVoucher.GetTypeName <> 'com.amazon.drm.Voucher@1.0') then
    raise Exception.Create('Unknown type, expected Voucher');

  FVoucher.StepIn;
  while FVoucher.HasNext do
  begin
    FVoucher.Next;
    if FVoucher.GetFieldName = 'cipher_iv' then
    begin
      FCipherIV := FVoucher.LOBValue;
    end
    else if FVoucher.GetFieldName = 'cipher_text' then
    begin
      FCipherText := FVoucher.LOBValue;
    end;
  end;
end;

procedure TDrmIonVoucher.PrintEnvelope(List: TStrings);
begin
  FEnvelope.Print(List);
end;

procedure TDrmIonVoucher.PrintKey(List: TStrings);
begin
  if FVoucher = nil then
    Parse;
  if FDRMKey = nil then
    DecryptVoucher;

  FDRMKey.Print(List);
end;

procedure TDrmIonVoucher.PrintVoucher(List: TStrings);
begin
  if FVoucher = nil then
    Parse;

  FVoucher.Print(List);
end;

{ TDrmIon }

constructor TDrmIon.Create(AIonStream: TStream; AOnVoucherRequired: TOnVoucherRequired);
begin
  FIon := TBinaryIonParser.Create(AIonStream);
  AddProtTable(FIon);
  FOnVoucherRequired := AOnVoucherRequired;
end;

destructor TDrmIon.Destroy;
begin
  FVoucher.Free;
  FIon.Free;

  inherited;
end;

procedure TDrmIon.Parse(const OutFN: string);
var
  CT, CIV: TBytes;
  OutPages: TFileStream;
begin
  FIon.Reset;
  if not FIon.HasNext then
    raise Exception.Create('DRMION envelope is empty');

  if (FIon.Next <> TID_SYMBOL) or (FIon.GetTypeName <> 'doctype') then
    raise Exception.Create('Expected doctype symbol');

  if (FIon.Next <> TID_LIST) or (FIon.GetTypeName <> 'com.amazon.drm.Envelope@2.0') then
    raise Exception.Create('Unknown type encountered in DRMION envelope, expected Envelope, got ' + FIon.GetTypeName);

  OutPages := TFileStream.Create(OutFN, fmCreate);
  repeat
    if FIon.GetTypeName = 'enddoc' then
      Break;

    FIon.StepIn;

    while FIon.HasNext do
    begin
      FIon.Next;
      if FIon.GetTypeName = 'com.amazon.drm.EnvelopeMetadata@2.0' then
      begin
        FIon.StepIn;
        while FIon.HasNext do
        begin
          FIon.Next;
          if FIon.GetFieldName = 'encryption_voucher' then
          begin
            if FVoucherName = '' then
            begin
              FVoucherName := FIon.StringValue;
              FVoucher := FOnVoucherRequired(FVoucherName);
              FVoucher.Parse;
              FVoucher.DecryptVoucher;
              FKey := FVoucher.SecretKey;
              if FKey = nil then
                raise Exception.Create('Unable to obtain secret key from voucher');
            end
            else if FVoucherName <> FIon.StringValue then
              raise Exception.Create('Unexpected: Different vouchers required for same file?');
          end;
        end;
        FIon.StepOut;
      end
      else if FIon.GetTypeName = 'com.amazon.drm.EncryptedPage@2.0' then
      begin
        FIon.StepIn;
        CT := nil;
        CIV := nil;
        while FIon.HasNext do
        begin
          FIon.Next;
          if FIon.GetFieldName = 'cipher_text' then
            CT := FIon.LOBValue
          else if FIon.GetFieldName = 'cipher_iv' then
            CIV := FIon.LOBValue;
        end;
        if (CT <> nil) and (CIV <> nil) then
          ProcessPage(CT, CIV, OutPages);
        FIon.StepOut;
      end;
    end;

    FIon.StepOut;
    if not FIon.HasNext then
      Break;
    FIon.Next;
  until False;
  OutPages.Free;
end;

procedure TDrmIon.Print(List: TStrings);
begin
  FIon.Print(List);
end;

procedure TDrmIon.ProcessPage(var CT, CIV: TBytes; OutPages: TStream);
var
  Comp: TBytesStream;
  Decomp: TMemoryStream;
  Props: array[0..4] of Byte;
  OutSize: UInt64;
begin
  with TCipher_Rijndael.Create do
  begin
    Mode := cmCBCx;
    Init(FKey[0], 16, CIV[0], 16);
    Decode(CT[0], CT[0], Length(CT));
    Free;
  end;
  RemovePKCS5Padding(CT);

  Decomp := TMemoryStream.Create;
  Comp := TBytesStream.Create(CT);
  try
    if ReadByte(Comp) <> 0 then
      raise Exception.Create('LZMA UseFilter not supported');

    with TLZMADecoder.Create do
    begin
      try
        Comp.Read(Props, 5);
        SetDecoderProperties(Props);
        Comp.Read(OutSize, 8);

        Code(Comp, Decomp, OutSize);
      finally
        Free;
      end;
    end;

    OutPages.CopyFrom(Decomp, 0);
  finally
    Comp.Free;
    Decomp.Free;
  end;
end;

end.
